#!/bin/bash

for run in HYDRO_FIDUCIAL HYDRO_FIDUCIAL_HiResDM HYDRO_JETS_published HYDRO_LOW_SIGMA8 HYDRO_LOW_SIGMA8_STRONGEST_AGN HYDRO_PLANCK HYDRO_PLANCK_DCDM_12 HYDRO_PLANCK_DCDM_24 HYDRO_PLANCK_LARGE_NU_FIXED HYDRO_PLANCK_LARGE_NU_VARY HYDRO_PLANCK_LARGER_NU_FIXED HYDRO_STRONG_AGN HYDRO_STRONGER_AGN HYDRO_STRONGER_AGN_STRONG_SUPERNOVA HYDRO_STRONGEST_AGN HYDRO_STRONG_JETS_published HYDRO_STRONG_SUPERNOVA HYDRO_WEAK_AGN ; do

    sbatch -J ${run} --array=0-1%1 ./sort_L1000N1800_bh_only.sh
    
done
